/*
 Copyright (C) 2010, 2011 Nicholas Peterson. All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
 * Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.
 
 * Neither the name of the author nor the names of its contributors may be used
 to endorse or promote products derived from this software without specific
 prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import <Foundation/Foundation.h>
#import "RPURLConnection.h"
#import "SBJson.h"

#define DEBUG_SERVER 1

#define RPAsyncServerErrorDomain @"RPAsyncServerErrorDomain"
#define RPAsyncServerKeyURL @"url"
#define RPAsyncServerKeyTag @"tag"
#define RPAsyncServerKeyDate @"date"



enum RPAsyncServerMethod{
    RPAsyncServerMethodGet = 0,
    RPAsyncServerMethodPost = 1,
    RPAsyncServerMethodPut = 2,
    RPAsyncServerMethodDelete = 3
};
typedef int RPAsyncServerMethod;
@protocol RPAsyncServerDelegate;
@class RPURLConnection;

/**
 @brief Creates, tracks, and dispatches server connections.
 
 Generates and tracks instances of RPURLConnection (NSURLConnection subclass). On completion the manager parses data into NSDictionary, looks for errors, and forwards connection results to the connections consumer. See RPURLConnection for more.
 
 **/
@interface RPAsyncServer : NSObject {
id <RPAsyncServerDelegate> delegate; //!< Consumer delegate
	NSMutableDictionary *connectionData; //!< tracks connections and their data.
    NSInteger historyCount; //!< count of total connections done while client has been running
	NSInteger bytesCount; //!< count of total bytes sent/recieved while client has been running

}
@property (nonatomic, assign) id <RPAsyncServerDelegate> delegate;
@property (nonatomic, retain) NSMutableDictionary *connectionData;
@property (nonatomic)         NSInteger historyCount; //for stat tracking; # of connections
@property (nonatomic)		  NSInteger bytesCount; // for stat tracking; amount of data transfer



/**
 @brief Singleton Allocation.
 
 Returns the current server object.
 \since v0.01.2

*/
+ (id)sharedServer;

#pragma mark - 
#pragma mark connection creation.
/**
 \section Subclass Overrides
 @brief Returns the base URL.
 
 Returns the base URL for all of the endpoints covered by this subclass. Defaults to 'example.com'
 \since v0.03
 */
- (NSString *)baseURL;

/**
@brief Returns a parsed and processed data structure. 

This is an optional override, By default the server assumes data is in JSON format. For additional processing call super and then process. 
return and NSError to report errors to consumer.
@param data Data returned by connection.

\since v0.03
*/

-(id)parseRecievedData:(NSData *)data;
/**
 \section Connection Creation
 @brief Generates Request.
 
Generates a NSURLRequest to be passed to an RPURLConnection.
 @param endpoint String to use for endpoint
 @param method RPAsyncServerMethod
 @param query Dictionary to set key = value querys
 @param headers Dictionary to set additional headers
 \since v0.03
 \see \li createConnectionWithRequest:tag:consumer: \li callServerWithAction:method:properties:consumer:
*/
-(NSURLRequest *)createRequestWithEndpoint:(NSString *)endpoint method:(RPAsyncServerMethod)method query:(NSDictionary *)query headers:(NSDictionary *)headers;

/**
 @brief Creates a new connection.
 
 creates and starts a new RPURLConnection.
 @param request An NSURLRequest. 
 @param tag String that will be used by the server object and the consumer to discriminate between connections, this should include the connection method and action/endpoint.
 @param consumer The consumer of the connection, must conform to the RPAsyncServerDelegate protocal
 \since v0.02
 \see \li createRequestWithEndpoint:method:query:headers: \li - (void)callServerWithTag:endpoint:method:query:headers:consumer:
 \section Notes
 \li Once the connection is created it will start immediately. 
 */
- (void)createConnectionWithRequest:(NSURLRequest *)request tag:(NSString *)tag consumer:(id)consumer;
/** 
 @brief Reports error to consumer
 
 Amends errors with connection details and date and forwards it to the consumer.
 
 @param error Original error  
 @param connection Connection that encountered error
 \since v0.04
 **/

- (void)sendError:(NSError *)error forConnection:(RPURLConnection *)connection;
/** 
 @brief Creates Server Call 
 
 Create the Request and Connection, Starts and adds the connection. This is usually called from within an access point.
 @param endpoint String this specific action (typically used by consumers to tell which connection is calling back).  
 @param endpoint String representing the endpoint on the server.  
 @param method Specify Get,Post,Put,or Delete. Defaults to RPAsyncServerMethodGet
 @param properties Dictionary used to set the HTML_BODY of the request. key = postkey and value = postval. \b Must not be nil.
 @param consumer The consumer of the connection, must conform to the RPAsyncServerDelegate protocal
 \since v0.03
 \see \li createRequestWithEndpoint:method:query:headers:
 **/
- (void)callServerWithTag:(NSString *)tag endpoint:(NSString *)endpoint method:(RPAsyncServerMethod)method 
                    query:(NSDictionary *)query headers:(NSDictionary *)headers consumer:(id)consumer;

/** 
 @brief Tests for similar connections 
 
 @param tag String representing the connection in the #connectionData array.  
 
 \since v0.02
 \see \li createConnectionWithRequest:tag:consumer: \li requestWithAction:method:properties:
 **/
- (BOOL)connectionInProgress:(NSString*)tag;
@end





/** 
 @brief Manages delegate call backs to consumer
 **/
@protocol RPAsyncServerDelegate <NSObject>

/** 
 @brief Responds to consumer with parsed information as dictionary.
 
 @param connection The calling connection. 
 @param info Dictionary with fully parsed response.
 \since v0.05
 **/
- (void)rpConnection:(RPURLConnection *)connection didFinishWithInfo:(NSDictionary *)info;

@optional
/** 
 @brief Responds to consumer in event of connection failure.
 
 While optional implementation of this method is highly advised.
 
 @param connection The calling connection. 
 @param error Error reported by connection, endpoint, or parser
 \since v0.05
 **/
- (void)rpConnection:(RPURLConnection *)connection didFailWithError:(NSError *)error;

/** 
 @brief Notifies consumer of connection response.
  
 @param connection The calling connection. 
 @param response NSURLResponse containing responce data.
 \since v0.05
 **/
- (void)rpConnection:(RPURLConnection *)connection didReceiveResponse:(NSURLResponse *)response;

/** 
 @brief Allows consumers to kill a connection.
 
 @param connection The calling connection. 
 \since v0.06
 **/
- (BOOL)rpConnectionShouldContinue:(RPURLConnection *)connection;

/** 
 @brief Responds to consumer with non-dictionary object.
 
 @param connection The calling connection. 
 @param info Dictionary with fully parsed response.
 \since v0.05
 **/
- (void)rpConnection:(RPURLConnection *)connection didFinishWithObject:(id)object;
@end
